#include <stdio.h>
#include <stdlib.h>

/* parameters of WELL19937a */
#define W_SIZE 32
#define N 624
#define P 31
#define MASKU (0xffffffffU>>(W_SIZE-P))
#define MASKL (~MASKU)
#define M1 70
#define M2 179
#define M3 449

#define MAT0POS(t,v) (v^(v>>t))
#define MAT0NEG(t,v) (v^(v<<(-(t))))
#define MAT1(v) v
#define MAT3POS(t,v) (v>>t)

#define V0            s->s[s->ptr]
#define VM1Over       s->s[s->ptr+M1-N]
#define VM1           s->s[s->ptr+M1]
#define VM2Over       s->s[s->ptr+M2-N]
#define VM2           s->s[s->ptr+M2]
#define VM3Over       s->s[s->ptr+M3-N]
#define VM3           s->s[s->ptr+M3]
#define VRm1          s->s[s->ptr-1]
#define VRm1Under     s->s[s->ptr+N-1]
#define VRm2          s->s[s->ptr-2]
#define VRm2Under     s->s[s->ptr+N-2]

#define newV0         s->s[s->ptr-1]
#define newV0Under    s->s[s->ptr-1+N]
#define newV1         s->s[s->ptr]
#define newVRm1       s->s[s->ptr-2]
#define newVRm1Under  s->s[s->ptr-2+N]

/* parameters for computing Jump */
#define MEXP 19937
#define P_SIZE ((MEXP/W_SIZE)+1)
#define LSB 0x00000001ul
#define QQ 8
#define LL 256 /* LL = 2^(QQ) */

/* the structure State represents states */
/* ptr indicates the leading term */
typedef struct{
  unsigned long s[N];
  int ptr; 
}State;

/* operations for polynomials */
void random_poly(unsigned long *);
void gray_code(void);
unsigned long get_coef(unsigned long *, unsigned int);
void set_coef(unsigned long *, unsigned int, unsigned long);

/* operations for states */
State *calc_state(unsigned long *, State *);
void copy_state(State *, State *);
State *horner1(unsigned long *, State *);
void gen_next(State *);
void add_state(State *, State *);
int compare_state(State *, State *);
void gen_vec_h(State *);

/* initial state */
State s0;

unsigned long h[LL];
State vec_h[LL];

/* these variable is used in next state computation of WELL 19937 */
static unsigned long z0, z1, z2;

void gen_next(State *s)
{
  if (s->ptr == 0){
    z0 = (VRm1Under & MASKL) | (VRm2Under & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1);
    z2 = MAT3POS (9, VM2) ^ MAT0POS (1, VM3);
    newV1      = z1 ^ z2;
    newV0Under = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr = N - 1;
  }
  
  else if (s->ptr == 1){
    z0 = (VRm1 & MASKL) | (VRm2Under & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1);
    z2 = MAT3POS (9, VM2) ^ MAT0POS (1, VM3);
    newV1 = z1 ^ z2;
    newV0 = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr = 0;
  }

  else if (s->ptr+M1 >= N){
    z0 = (VRm1 & MASKL) | (VRm2 & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1Over);
    z2 = MAT3POS (9, VM2Over) ^ MAT0POS (1, VM3Over);
    newV1 = z1 ^ z2;
    newV0 = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr--;
  }

  else if (s->ptr+M2 < N && s->ptr+M3 >= N){
    z0 = (VRm1 & MASKL) | (VRm2 & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1);
    z2 = MAT3POS (9, VM2) ^ MAT0POS (1, VM3Over);
    newV1 = z1 ^ z2;
    newV0 = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr--;
  }

  else if (s->ptr+M1 < N && s->ptr+M2 >= N){
    z0 = (VRm1 & MASKL) | (VRm2 & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1);
    z2 = MAT3POS (9, VM2Over) ^ MAT0POS (1, VM3Over);
    newV1 = z1 ^ z2;
    newV0 = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr--;
  }

  else if (s->ptr + M3 < N && 2 <= s->ptr){
    z0 = (VRm1 & MASKL) | (VRm2 & MASKU);
    z1 = MAT0NEG (-25, V0) ^ MAT0POS (27, VM1);
    z2 = MAT3POS (9, VM2) ^ MAT0POS (1, VM3);
    newV1 = z1 ^ z2;
    newV0 = MAT1 (z0) ^ MAT0NEG (-9, z1) ^ MAT0NEG (-21, z2) ^ MAT0POS (21, newV1);
    s->ptr--;
  }
}
int main(void)
{
  unsigned long *pf;
  State *ss1, *ss2;
  int i, j;
  char c;
  FILE *fin;

  pf = (unsigned long *)calloc(P_SIZE, sizeof(unsigned long));

  /* read the file clist.txt, and set the coefficient */
  if ((fin = fopen("clist_well19937.txt","r")) == NULL){
    printf("File read error.\n");
    exit(1);
  }
  
  for (i=MEXP-1; i>-1; i--){
    c = fgetc(fin);
    if (c == '1')
      set_coef(pf, i, 1);
  }
    
  /* computes jumping ahead with standard Horner method */
  ss1 = horner1(pf, &s0);
  
  /* computes jumping ahead with Sliding window algorithm */
  gen_vec_h(&s0);
  ss2 = calc_state(pf, &s0);
  
  if (compare_state(ss1, ss2) != 0)
    printf("error\n");

  return 0;
}

/* generate random polynomial */
void random_poly(unsigned long *p)
{
  int i;

  for (i=0; i<MEXP; i++)
      set_coef(p, i, rand() % 2);
}

/* copy state ss */
void copy_state(State *ts, State *ss)
{
  int i;

  for(i=0; i<N; i++)
    ts->s[i] = ss->s[i];

  ts->ptr = ss->ptr;
}

/* this function computes jumping-ahead */
/* using standard Horner method         */
State *horner1(unsigned long *pf, State *ss)
{
  int i=MEXP-1;
  State *temp;

  temp = (State *)calloc(1, sizeof(State));

  while(get_coef(pf,i) == 0)
    i--;

  if (i != 0){
    copy_state(temp, ss);
    gen_next(temp);
    i--;
    for (; i>0; i--){
      if (get_coef(pf,i) == 1)
	add_state(temp, ss);
      else
	;
      gen_next(temp);
    }
    if (get_coef(pf,0) == 1)
      add_state(temp, ss);
    else
      ;
  }
  else
    ;

  return(temp);
}

/* this function computes addtion in state space */
void add_state(State *s1, State *s2) 
{
  int i, pt1=s1->ptr, pt2=s2->ptr;
  
  if (pt2-pt1 >= 0){
    for (i=0; i<N-pt2; i++)
      s1->s[i+pt1] ^= s2->s[i+pt2];
    for(; i<N-pt1; i++)
      s1->s[i+pt1] ^= s2->s[i+(pt2-N)];
    for(; i<N; i++)
      s1->s[i+(pt1-N)] ^= s2->s[i+(pt2-N)];
  }
  else{
    for (i=0; i<N-pt1; i++)
      s1->s[i+pt1] ^= s2->s[i+pt2];
    for(; i<N-pt2; i++)
      s1->s[i+(pt1-N)] ^= s2->s[i+pt2];
    for(; i<N; i++)
      s1->s[i+(pt1-N)] ^= s2->s[i+(pt2-N)];
  }
}

/* this function computes jumping-ahead */
/* using Sliding window algorithm       */
State *calc_state(unsigned long *pf, State *ss)
{
  State *temp1;
  int i=MEXP-1, j, digit, skip=0;

  temp1 = (State *)calloc(1, sizeof(State));

  while (get_coef(pf,i) == 0)
    i--;

  for (; i>=QQ; i--){
    if (get_coef(pf,i) != 0){
      for (j=0; j<QQ+1; j++)
	gen_next(temp1);
      digit = 0; 
      for (j=0; j<QQ; j++)
	digit = (digit << 1) ^ get_coef(pf, i-j-1);
      add_state(temp1, &vec_h[digit]);
      i -= QQ;
    }
    else
      gen_next(temp1);
  }
  
  for (; i>-1; i--){
    gen_next(temp1);
    if (get_coef(pf,i) == 1)
      add_state(temp1, ss);
    else
      ;
  }
  return(temp1);
}

/* compare s1 with s2 */
/* if s1 is different from s2, then return 1 , else return 0 */
int compare_state(State *s1, State *s2)
{
  int i, d = 0;

  for (i=0; i<N-1; i++)
    if (s1->s[(i+s1->ptr)%N] != s2->s[(i+s2->ptr)%N]){
      d = 1;
      break;
    }
  if ((s1->s[(N-1+s1->ptr)%N] & 0x80000000ul) != (s2->s[(N-1+s2->ptr)%N] & 0x80000000ul))
    d = 1;

  return(d);
}

/* generate Gray code */
void gray_code(void)
{
  unsigned int i, j=1, l=1, term = LL;

  h[0] = 0;
  
  for (i=1; i<=QQ; i++){
      l = (l << 1);
      term = (term >> 1);
      for (; j<l; j++)
	  h[j] = h[l-j-1] ^ term; 
  }
}

/* computes h(f)s, for all q-degree polynomials,            */
/* f is the tarnsition function, and s is the initial state */
/* the results are stored in vec_h[0] , ... , vec_h[LL-1]   */
void gen_vec_h(State *s)
{
  int i;
  unsigned long k, g;
  State v;

  gray_code();

  copy_state(&vec_h[0], s);

  for(i=0; i<QQ; i++)
    gen_next(&vec_h[0]);

  for (i=1; i<LL; i++){
    copy_state(&v, s);
    g = h[i] ^ h[i-1];
    for (k=1; k<g; k=(k<<1))
      gen_next(&v);
    copy_state(&vec_h[h[i]], &vec_h[h[i-1]]);
    add_state(&vec_h[h[i]], &v);
  }
}


/* 32-bits function */ 
/* returns the i-th coefficient of the polynomial pf */
unsigned long get_coef(unsigned long *pf, unsigned int deg)
{
  if ((pf[deg >> 5] & (LSB << (deg & 0x1ful))) != 0) 
    return(1);
  else
    return(0);
}

/* 32-bit function */
/* sets the coefficient of the polynomial pf to v */
void set_coef(unsigned long *pf, unsigned int deg, unsigned long v)
{
  if (v != 0)
    pf[deg >> 5] ^= (LSB << (deg & 0x1ful));
  else
    ;
}
